/*
 * abaddrd.c - address-range database server for abmail
 *
 * Copyright (C) 2008 Taiji Yamada <taiji@aihara.co.jp>
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Affero General Public License for more details.
 */
/*
 * usage:
 *	# ./abaddrd [-C] [-R] [-W] [-w wait_sec] [-f path_to.bdb]
 */
#include <stdio.h>	/* stderr, etc */
#include <stdlib.h>	/* qsort, bsearch, etc */
#include <string.h>	/* strcmp, etc */
#include <libgen.h>	/* dirname, etc */
#include <errno.h>	/* errno */
#include <syslog.h>	/* LOG_ERR, etc */
#if defined(USE_TI_RPC)
#include <netdir.h>
#include <netconfig.h>
#endif
#include <unistd.h>	/* gethostname, etc */
#include <sys/types.h>	/* netinet/in.h, arpa/inet.h, sys/socket.h, etc */
#include <netinet/in.h>	/* in_addr, etc */
#include <arpa/inet.h>	/* inet_ntoa, etc */
#include <sys/socket.h>	/* AF_INET, etc */
#include <netdb.h>	/* gethostbyname, etc */
#include <time.h>	/* time_t, etc */
#include <sys/time.h>	/* struct timeval, etc */
#include <signal.h>	/* SIGTERM, etc */
#if defined(USE_QDBM)
#include <villa.h>
#else
#include <tcbdb.h>
#endif
#include "abcidr.h"
#include "abmisc.h"
#include "abaddr.h"

#define min(a, b) (((a)<(b))?(a):(b))
#define max(a, b) (((a)>(b))?(a):(b))
#define lbu(l, b, u) min(max((l), (b)), (u))

#if !defined(HAVE_STRLCPY)
#define strlcpy(d, s, z) do { strncpy(d, s, z); if ((z) > 0) (d)[(z)-1] = '\0'; } while (0)
#endif

#if !defined(HAVE_STRLCAT)
#define strlcat(d, s, z) do { if ((z) > strlen(d)) strncat(d, s, (z)-strlen(d)-1); } while (0)
#endif

#if !defined(HAVE_INET_ATON)
#define inet_aton(s, p) ((((p)->s_addr=inet_addr(s))!=-1||strcmp(s, "255.255.255.255")==0)?1:0)
#endif

#if !defined(PREFIX)
#define PREFIX "/opt/local"
#endif

struct abaddr_db {
  const char *pathname;
  int clear, read_lst, write_lst;
  long wt;	/* wait [sec] */
  int argc; char **argv;
#if defined(USE_QDBM)
  VILLA *bdb;
#else
  TCBDB *bdb;
#endif
} me = {
  "./abaddr.bdb",
  0, 0, 0,
  60*60*24*7,
  0, NULL,
};

#if defined(USE_QDBM)
int abaddr_bdbcmp(const char *a, int az, const char *b, int bz)
#else
int abaddr_bdbcmp(const char *a, int az, const char *b, int bz, void *op)
#endif
{
  int a_af = AF_UNSPEC, b_af = AF_UNSPEC;
  char *a_ads[2] = { (char *)a, }, *b_ads[2] = { (char *)b, };

  switch (az) {
  case sizeof(struct in_addr)*2:
    a_af = AF_INET;
    a_ads[1] = &a_ads[0][4];
    break;
#if !defined(DISABLE_IPv6)
  case sizeof(struct in6_addr)*2:
    a_af = AF_INET6;
    a_ads[1] = &a_ads[0][16];
    break;
#endif
  }
  switch (bz) {
  case sizeof(struct in_addr)*2:
    b_af = AF_INET;
    b_ads[1] = &b_ads[0][4];
    break;
#if !defined(DISABLE_IPv6)
  case sizeof(struct in6_addr)*2:
    b_af = AF_INET6;
    b_ads[1] = &b_ads[0][16];
    break;
#endif
  }
  if (a_af != b_af)
    return (a_af<b_af)?-1:1;
  else if (a_af == AF_INET) {
    if (ntohl(((struct in_addr *)b_ads[0])->s_addr) <= ntohl(((struct in_addr *)a_ads[1])->s_addr) &&
	ntohl(((struct in_addr *)a_ads[0])->s_addr) <= ntohl(((struct in_addr *)b_ads[1])->s_addr))
      return 0;
    else if (ntohl(((struct in_addr *)a_ads[1])->s_addr) < ntohl(((struct in_addr *)b_ads[0])->s_addr))
      return -1;
    else
      return 1;
  }
#if !defined(DISABLE_IPv6)
  else if (a_af == AF_INET6) {
    if (in6_addr_le(*(struct in6_addr *)b_ads[0], *(struct in6_addr *)a_ads[1]) &&
	in6_addr_le(*(struct in6_addr *)a_ads[0], *(struct in6_addr *)b_ads[1]))
      return 0;
    else if (in6_addr_lt(*(struct in6_addr *)a_ads[1], *(struct in6_addr *)b_ads[0]))
      return -1;
    else
      return 1;
  }
#endif
  return 1;
}

struct abaddrat {
  int af;
  union {
    struct in_addr in;
#if !defined(DISABLE_IPv6)
    struct in6_addr in6;
#endif
  } ads[2];
  int rc;	/* related code */
  long lt;	/* life time [sec], 0: infinite, -1: immediate, -2: conditional */
  char c[128];	/* characters */
};

void am_abaddrat(struct am *am, struct abaddrat *at)
{
  at->af = AF_INET;
  at->ads[0].in = am->ads[2];
  at->ads[1].in = am->ads[3];
  at->rc = am->rc;
  at->lt = 0;
  strlcpy(at->c, am->c?am->c:"", sizeof(at->c));
}
#if !defined(DISABLE_IPv6)
void am6_abaddrat(struct am6 *am, struct abaddrat *at)
{
  at->af = AF_INET6;
  at->ads[0].in6 = am->ads[2];
  at->ads[1].in6 = am->ads[3];
  at->rc = am->rc;
  at->lt = 0;
  strlcpy(at->c, am->c?am->c:"", sizeof(at->c));
}
#endif

void abaddrat_fprint(struct abaddrat *at, FILE *fp)
{
  char buf[3][1024];

  inet_ntop(at->af, &at->ads[0], buf[0], sizeof(buf[0]));
  inet_ntop(at->af, &at->ads[1], buf[1], sizeof(buf[1]));
  strftime(buf[2], sizeof(buf[2]), "%Y/%m/%d %H:%M:%S", localtime((time_t *)&at->lt));
  fprintf(fp, "%s-%s\t%d\t%s\t%s\n", buf[0], buf[1], at->rc,
	  (at->lt>0)?buf[2]:"----/--/-- --:--:--", at->c);
  fflush(stderr);
}

size_t abaddrat_make_key(struct abaddrat *at, char *b, size_t z)
{
  size_t s = 0;

  switch (at->af) {
  case AF_INET:
    if (sizeof(struct in_addr)*2 <= z) {
      memcpy(&b[0], &at->ads[0].in, sizeof(struct in_addr));
      memcpy(&b[4], &at->ads[1].in, sizeof(struct in_addr));
      s = sizeof(struct in_addr)*2;
    }
    break;
#if !defined(DISABLE_IPv6)
  case AF_INET6:
    if (sizeof(struct in6_addr)*2 <= z) {
      memcpy(&b[0], &at->ads[0].in6, sizeof(struct in6_addr));
      memcpy(&b[16], &at->ads[1].in6, sizeof(struct in6_addr));
      s = sizeof(struct in6_addr)*2;
    }
    break;
#endif
  }
  return s;
}
size_t abaddrat_make_val(struct abaddrat *at, char *b, size_t z)
{
  size_t s = 0;
  int8_t rc = at->rc;
  int32_t lt = htonl((at->lt>0)?time(NULL)+at->lt:at->lt);

  at->lt = lt;	/* ! */
  if (sizeof(int8_t)+sizeof(int32_t) <= z) {
    memcpy(&b[0], &rc, sizeof(int8_t));
    memcpy(&b[1], &lt, sizeof(int32_t));
    s = sizeof(int8_t)+sizeof(int32_t);
    strlcpy(&b[5], at->c, z-s);
    s += strlen(&b[5]);
  }
  return s;
}
void abaddrat_make_key_val(struct abaddrat *at, char *kp, size_t kz, size_t *ks, char *vp, size_t vz, size_t *vs)
{
  *ks = abaddrat_make_key(at, kp, kz);
  *vs = abaddrat_make_val(at, vp, vz);
#ifdef DEBUG
  fprintf(stderr, "requested addr:\t");
  abaddrat_fprint(at, stderr);
#endif
}
void abaddrat_read_key(struct abaddrat *at, char *b, size_t z)
{
  switch (z) {
  case sizeof(struct in_addr)*2:
    at->af = AF_INET;
    memcpy(&at->ads[0].in, &b[0], sizeof(struct in_addr));
    memcpy(&at->ads[1].in, &b[4], sizeof(struct in_addr));
    break;
#if !defined(DISABLE_IPv6)
  case sizeof(struct in6_addr)*2:
    at->af = AF_INET6;
    memcpy(&at->ads[0].in6, &b[0], sizeof(struct in6_addr));
    memcpy(&at->ads[1].in6, &b[16], sizeof(struct in6_addr));
    break;
#endif
  }
}
void abaddrat_read_val(struct abaddrat *at, char *b, size_t z)
{
  int8_t rc;
  int32_t lt;

  if (sizeof(int8_t)+sizeof(int32_t) <= z) {
    memcpy(&rc, &b[0], sizeof(int8_t));
    memcpy(&lt, &b[1], sizeof(int32_t));
    at->rc = rc;
    at->lt = ntohl(lt);
    if (sizeof(int8_t)+sizeof(int32_t) < z)
      strlcpy(at->c, &b[5], sizeof(at->c));
  }
}
void abaddrat_read_key_val(struct abaddrat *at, char *kp, size_t kz, char *vp, size_t vz)
{
  abaddrat_read_key(at, kp, kz);
  abaddrat_read_val(at, vp, vz);
}

int abaddrat_db_add(struct abaddr_db *me, struct abaddrat *at)
{
  int rv = !0;
  char kb[1024], vb[1024];
  size_t kz, vz;

  abaddrat_make_key_val(at, kb, sizeof(kb), &kz, vb, sizeof(vb), &vz);
#if defined(USE_QDBM)
  if (!vlput(me->bdb, kb, kz, vb, vz, VL_DOVER)) {
    syslog(LOG_ERR, "vlput failed: %s", dperrmsg(dpecode));
    rv = 0;
  }
  /*vlsync(me->bdb);*/
#else
  if (!tcbdbput(me->bdb, kb, kz, vb, vz)) {
    syslog(LOG_ERR, "tcbdbput failed: %s", tcbdberrmsg(tcbdbecode(me->bdb)));
    rv = 0;
  }
  /*tcbdbsync(me->bdb);*/
#endif
  return rv;
}

int abaddrat_db_del(struct abaddr_db *me, struct abaddrat *at)
{
  /*
    database
    -1              1    0 1
    +------+        +--+ + +------+
    add request	!(lt<0)
        -1
        +----------------------+
    result
    -1  -1                      1
    +--++----------------------++-+

    database
    -1              1    0 1
    +------+        +--+ + +------+
    immediately del request	(lt==-1)
        -1
        +----------------------+
    result
    -1                          1
    +--+                        +-+

    database
    -1              1    0 1
    +------+        +--+ + +------+
    conditionally del request	(lt==-2)
        -1
        +----------------------+
    result
    -1              1    0 1
    +--+            +--+ + +------+
   */
  int rv = !0;
  struct abaddrat ab[] = { *at, *at };
  struct abaddrat ac[] = { { AF_UNSPEC, }, { AF_UNSPEC, } };
  char kb[2][1024];
  size_t kz[2];
  int i;
#if defined(USE_QDBM)

  ab[0].ads[1] = ab[0].ads[0];
  ab[1].ads[0] = ab[1].ads[1];
  kz[0] = abaddrat_make_key(&ab[0], kb[0], sizeof(kb[0]));
  kz[1] = abaddrat_make_key(&ab[1], kb[1], sizeof(kb[1]));
  vlcurjump(me->bdb, kb[0], kz[0], VL_JFORWARD);
  {
    char *ckp, *cvp;
    int ckz, cvz;

    while ((ckp=vlcurkey(me->bdb, &ckz))) {
      struct abaddrat at = { 0, };

      if (abaddr_bdbcmp(ckp, ckz, kb[1], kz[1]) == 1) {
	free(ckp);
	break;
      }
      cvp = vlcurval(me->bdb, &cvz);
      abaddrat_read_key_val(&at, ckp, ckz, cvp, cvz);
      free(ckp);
      free(cvp);
      if (!(!(ab[0].lt<0)||ab[0].lt==-1||(ab[0].lt==-2&&ab[0].rc==at.rc))) {
	vlcurnext(me->bdb);
	continue;
      }
      switch (at.af) {
      case AF_INET:
	if (ntohl(at.ads[0].in.s_addr) < ntohl(ab[0].ads[0].in.s_addr)) {
	  ac[0] = at;
	  ac[0].ads[1].in.s_addr = htonl(ntohl(ab[0].ads[0].in.s_addr)-1);
	}
	if (ntohl(ab[1].ads[1].in.s_addr) < ntohl(at.ads[1].in.s_addr)) {
	  ac[1] = at;
	  ac[1].ads[0].in.s_addr = htonl(ntohl(ab[1].ads[1].in.s_addr)+1);
	}
	break;
#if !defined(DISABLE_IPv6)
      case AF_INET6:
	if (in6_addr_lt(at.ads[0].in6, ab[0].ads[0].in6)) {
	  ac[0] = at;
	  ac[0].ads[1].in6 = in6_addr_sub(ab[0].ads[0].in6, in6_addr_one());
	}
	if (in6_addr_lt(ab[1].ads[1].in6, at.ads[1].in6)) {
	  ac[1] = at;
	  ac[1].ads[0].in6 = in6_addr_add(ab[1].ads[1].in6, in6_addr_one());
	}
	break;
#endif
      }
      rv = vlcurout(me->bdb);
    }
  }
  /*vlsync(me->bdb);*/
#else
  TCLIST *lst;

  ab[0].ads[1] = ab[0].ads[0];
  ab[1].ads[0] = ab[1].ads[1];
  kz[0] = abaddrat_make_key(&ab[0], kb[0], sizeof(kb[0]));
  kz[1] = abaddrat_make_key(&ab[1], kb[1], sizeof(kb[1]));
  if ((lst=tcbdbrange(me->bdb, kb[0], kz[0], !0, kb[1], kz[1], !0, -1))) {
    for (i=0; i<tclistnum(lst); i++) {
      char *kp, *vp;
      int kz, vz;
      struct abaddrat at = { 0, };

      kp = (char *)tclistval(lst, i, &kz);
      vp = (char *)tcbdbget(me->bdb, kp, kz, &vz);
      abaddrat_read_key_val(&at, kp, kz, vp, vz);
      free(vp);
      if (!(!(ab[0].lt<0)||ab[0].lt==-1||(ab[0].lt==-2&&ab[0].rc==at.rc)))
	continue;
      switch (at.af) {
      case AF_INET:
	if (ntohl(at.ads[0].in.s_addr) < ntohl(ab[0].ads[0].in.s_addr)) {
	  ac[0] = at;
	  ac[0].ads[1].in.s_addr = htonl(ntohl(ab[0].ads[0].in.s_addr)-1);
	}
	if (ntohl(ab[1].ads[1].in.s_addr) < ntohl(at.ads[1].in.s_addr)) {
	  ac[1] = at;
	  ac[1].ads[0].in.s_addr = htonl(ntohl(ab[1].ads[1].in.s_addr)+1);
	}
	break;
#if !defined(DISABLE_IPv6)
      case AF_INET6:
	if (in6_addr_lt(at.ads[0].in6, ab[0].ads[0].in6)) {
	  ac[0] = at;
	  ac[0].ads[1].in6 = in6_addr_sub(ab[0].ads[0].in6, in6_addr_one());
	}
	if (in6_addr_lt(ab[1].ads[1].in6, at.ads[1].in6)) {
	  ac[1] = at;
	  ac[1].ads[0].in6 = in6_addr_add(ab[1].ads[1].in6, in6_addr_one());
	}
	break;
#endif
      }
      rv = tcbdbout(me->bdb, kp, kz);
    }
    tclistdel(lst);
  }
  /*tcbdbsync(me->bdb);*/
#endif
  for (i=0; i<2; i++)
    if (ac[i].af != AF_UNSPEC)
      rv = abaddrat_db_add(me, &ac[i]);
  return rv;
}

int abaddrat_db_get(struct abaddr_db *me, struct abaddrat *at)
{
  int rv = !0;
  char kb[1024], *vp;
  int kz, vz;

  kz = abaddrat_make_key(at, kb, sizeof(kb));
#if defined(USE_QDBM)
  if (!(vp=vlget(me->bdb, kb, kz, &vz)))
    rv = 0;
#else
  if (!(vp=(char *)tcbdbget(me->bdb, kb, kz, &vz)))
    rv = 0;
#endif
  if (rv)
    abaddrat_read_val(at, vp, vz);
  if (at->lt>0 && at->lt<time(NULL))
    rv = 0;
  return rv;
}

void abaddr_db_prepare(struct abaddr_db *me)
{
#if defined(USE_QDBM)
  if (!(me->bdb=vlopen(me->pathname, VL_OWRITER|VL_OCREAT|(me->clear?VL_OTRUNC:0), abaddr_bdbcmp))) {
    syslog(LOG_ERR, "vlopen failed: %s", dperrmsg(dpecode));
    exit(3);
  }
#else
  if (!((me->bdb=tcbdbnew()) &&
	tcbdbsetcmpfunc(me->bdb, abaddr_bdbcmp, NULL) &&
	tcbdbopen(me->bdb, me->pathname, BDBOWRITER|BDBOCREAT|(me->clear?BDBOTRUNC:0)))) {
    syslog(LOG_ERR, "tcbdbopen failed: %s", tcbdberrmsg(tcbdbecode(me->bdb)));
    exit(3);
  }
#endif
}

void abaddr_db_sync(struct abaddr_db *me)
{
#if defined(USE_QDBM)
  vlsync(me->bdb);
#else
  tcbdbsync(me->bdb);
#endif
}

void abaddr_db_terminate(struct abaddr_db *me)
{
  abaddr_db_sync(me);
#if defined(USE_QDBM)
  if (!vlclose(me->bdb))
    syslog(LOG_ERR, "vlclose failed: %s", dperrmsg(dpecode));
#else
  if (!tcbdbclose(me->bdb))
    syslog(LOG_ERR, "tcbdbclose failed: %s", tcbdberrmsg(tcbdbecode(me->bdb)));
  tcbdbdel(me->bdb);
#endif
}

void abaddr_db_amset_read(struct abaddr_db *me,
			  struct amset *as,
#if !defined(DISABLE_IPv6)
			  struct am6set *a6s,
#endif
			  int info, int check)
{
  int i;

  amset_read(as,
#if !defined(DISABLE_IPv6)
	     a6s,
#endif
	     info, check, me->argc, me->argv);
  for (i=0; i<as->nam; i++) {
    struct abaddrat at = { 0, };

    am_abaddrat(&as->ams[i], &at);
    abaddrat_db_del(me, &at);
    abaddrat_db_add(me, &at);
  }
  amset_free(as);
#if !defined(DISABLE_IPv6)
  for (i=0; i<a6s->nam; i++) {
    struct abaddrat at = { 0, };

    am6_abaddrat(&a6s->ams[i], &at);
    abaddrat_db_del(me, &at);
    abaddrat_db_add(me, &at);
  }
  am6set_free(a6s);
#endif
  abaddr_db_sync(me);
}

void abaddrat_am_fprintln(struct abaddrat *at, FILE *fp)
{
  struct am am = { "", "", 0, NULL, };
#if !defined(DISABLE_IPv6)
  struct am6 am6 = { "", "", 0, NULL, };
#endif

  switch (at->af) {
  case AF_INET:
    am.ads[0] = am.ads[2] = at->ads[0].in;
    am.ads[3] = at->ads[1].in;
    strlcpy(am.a, inet_ntoa(am.ads[0]), sizeof(am.a));
    if (am_fprint_cidrs(&am, NULL) == 1)
      am_fprint_cidrs(&am, fp);
    else
      am_fprint_range(&am, fp);
    break;
#if !defined(DISABLE_IPv6)
  case AF_INET6:
    am6.ads[0] = am6.ads[2] = at->ads[0].in6;
    am6.ads[3] = at->ads[1].in6;
    inet_ntop(AF_INET6, &am.ads[0], am.a, sizeof(am.a));
    if (am6_fprint_cidrs(&am6, NULL) == 1)
      am6_fprint_cidrs(&am6, fp);
    else
      am6_fprint_range(&am6, fp);
    break;
#endif
  }
  switch (at->af) {
  case AF_INET:
#if !defined(DISABLE_IPv6)
  case AF_INET6:
#endif
    fprintf(fp, "\t%s\n", at->c);
  }
}

void abaddr_db_amset_write(struct abaddr_db *me)
{
  char *myname = me->pathname?strdup(me->pathname):NULL, *herename = myname?dirname(myname):".";
  char lst_pathname[MAXPATHLEN];
  struct { char *basename; int rc; } lst[] = {
    { "ams_trusted.lst.out", -1 },
    { "ams_reliable.lst.out", 0 },
    { "ams_unreliable.lst.out", 1 },
  };
  int i;
  FILE *fp;

  for (i=0; i<sizeof(lst)/sizeof(lst[0]); i++) {
    sprintf(lst_pathname, "%s/%s", herename, lst[i].basename);
    if (!(fp=fopen(lst_pathname, "w")))
      continue;
    {
      char *kp, *vp;
      int kz, vz;
#if defined(USE_QDBM)

      vlcurfirst(me->bdb);
      while ((kp=vlcurkey(me->bdb, &kz))) {
	struct abaddrat at = { 0, };

	vp = vlcurval(me->bdb, &vz);
	abaddrat_read_key_val(&at, kp, kz, vp, vz);
	free(kp);
	free(vp);
	if (!(at.lt>0) && at.rc == lst[i].rc)
	  abaddrat_am_fprintln(&at, fp);
	vlcurnext(me->bdb);
      }
#else
      BDBCUR *cur;

      cur = tcbdbcurnew(me->bdb);
      tcbdbcurfirst(cur);
      while ((kp=tcbdbcurkey(cur, &kz))) {
	struct abaddrat at = { 0, };

	vp = tcbdbcurval(cur, &vz);
	abaddrat_read_key_val(&at, kp, kz, vp, vz);
	free(kp);
	free(vp);
	if (!(at.lt>0) && at.rc == lst[i].rc)
	  abaddrat_am_fprintln(&at, fp);
	tcbdbcurnext(cur);
      }
      tcbdbcurdel(cur);
#endif
    }
    fclose(fp);
  }
  free(myname);
}

void abaddr_db_cycle_life(struct abaddr_db *me)
{
  char *kp, *vp;
  int kz, vz;
#if defined(USE_QDBM)

#ifdef DEBUG
  fprintf(stderr, "num alive:\t%d\n", vlrnum(me->bdb));
#endif
  vlcurfirst(me->bdb);
  while ((kp=vlcurkey(me->bdb, &kz))) {
    struct abaddrat at = { 0, };

    vp = vlcurval(me->bdb, &vz);
    abaddrat_read_key_val(&at, kp, kz, vp, vz);
    free(kp);
    free(vp);
    if (at.lt>0 && at.lt<time(NULL)) {
#ifdef DEBUG
      fprintf(stderr, "expired addr:\t");
      abaddrat_fprint(&at, stderr);
#endif
      vlcurout(me->bdb);
      /*vlsync(me->bdb);*/
    }
    else {
#ifdef DEBUG
      fprintf(stderr, "alive addr:\t");
      abaddrat_fprint(&at, stderr);
#endif
    }
    vlcurnext(me->bdb);
  }
#else
  BDBCUR *cur;

#ifdef DEBUG
  fprintf(stderr, "num alive:\t%llu\n", tcbdbrnum(me->bdb));
#endif
  cur = tcbdbcurnew(me->bdb);
  tcbdbcurfirst(cur);
  while ((kp=tcbdbcurkey(cur, &kz))) {
    struct abaddrat at = { 0, };

    vp = tcbdbcurval(cur, &vz);
    abaddrat_read_key_val(&at, kp, kz, vp, vz);
    free(kp);
    free(vp);
    if (at.lt>0 && at.lt<time(NULL)) {
#ifdef DEBUG
      fprintf(stderr, "expired addr:\t");
      abaddrat_fprint(&at, stderr);
#endif
      tcbdbcurout(cur);
      /*tcbdbsync(me->bdb);*/
    }
    else {
#ifdef DEBUG
      fprintf(stderr, "alive addr:\t");
      abaddrat_fprint(&at, stderr);
#endif
    }
    tcbdbcurnext(cur);
  }
  tcbdbcurdel(cur);
#endif
  abaddr_db_sync(me);
}

static int is_sigterm = 0;
void handler_sigterm(int n)
{
  is_sigterm = !0;
}
int abaddr_svc_run(void)
{
  time_t awake;
  int nreadfd;
  fd_set readfds;
  struct timeval timeout = { me.wt, };

  signal(SIGTERM, handler_sigterm);
  awake = time(NULL) + me.wt;
  while (!0) {
    readfds = svc_fdset;
    nreadfd = select(FD_SETSIZE, &readfds, NULL, NULL, &timeout);
    switch (nreadfd) {
    case -1:
      if (errno != EINTR)
	syslog(LOG_ERR, "select failed: %m");
      else if (is_sigterm) {
	abaddr_db_terminate(&me);
	exit(0);
      }
      break;
    case 0:
      abaddr_db_cycle_life(&me);
      if (me.write_lst) abaddr_db_amset_write(&me);
      awake = time(NULL) + me.wt;
      break;
    default:
      if (!(time(NULL) < awake)) {
	awake = time(NULL) + me.wt;
	abaddr_db_cycle_life(&me);
	if (me.write_lst) abaddr_db_amset_write(&me);
      }
      svc_getreqset(&readfds);
    }
  }
  return 0;
}

abaddrproc_stat abaddrproc(struct svc_req *rqstp)
{
  abaddrproc_stat stat;
#if defined(USE_TI_RPC)
  struct netbuf *nb;
  struct netconfig *nc;
  char *uaddr, *p;
#else
  struct sockaddr_in *saddr;
#endif
  struct in_addr caddr;
#if !defined(DISABLE_IPv6)
  struct in6_addr c6addr;
  int af = AF_INET;
#define case_af(a6,a) ((af==AF_INET6)?(a6):(a))
#else
#define case_af(a6,a) (a)
#endif

  stat = ABADDRPROC_OK;
#if defined(USE_TI_RPC)
  if (!((nc=getnetconfigent(rqstp->rq_xprt->xp_netid)) &&
	(nb=svc_getrpccaller(rqstp->rq_xprt)) && (uaddr=taddr2uaddr(nc, nb)))) {
    if (nc) freenetconfigent(nc);
    stat = ABADDRPROC_EIO;
    return stat;
  }
  if ((p=strrchr(uaddr, '.'))) {
    *p = '\0';
    if ((p=strrchr(uaddr, '.')))
      *p = '\0';
  }
  if (inet_aton(uaddr, &caddr) == 0) {
#if !defined(DISABLE_IPv6)
    if (inet_pton(AF_INET6, uaddr, &c6addr) == 0) {
      stat = ABADDRPROC_EIO;
      return stat;
    }
    else
      af = AF_INET6;
#else
    stat = ABADDRPROC_EIO;
    return stat;
#endif
  }
  freenetconfigent(nc);
  free(uaddr);
#else
  if (!(saddr=svc_getcaller(rqstp->rq_xprt))) {
    stat = ABADDRPROC_EIO;
    return stat;
  }
  caddr.s_addr = saddr->sin_addr.s_addr;	/* TS_RPC supports IPv4 only */
#endif
#ifdef DEBUG
  {
#if !defined(DISABLE_IPv6)
    char buf[1024];
#endif

    fprintf(stderr, "client addr:\t%s\n",
	    case_af(inet_ntop(af, &c6addr, buf, sizeof(buf)),
		    inet_ntoa(caddr)));
    fflush(stderr);
  }
#endif
  if (!(case_af((void *)am6_search(am6set_clients.nam, am6set_clients.ams, c6addr),
		(void *)am_search(amset_clients.nam, amset_clients.ams, caddr)))) {
    stat = ABADDRPROC_EACCES;
    return stat;
  }
  return stat;
}

abaddrproc_stat abaddrproc_addr(struct abaddrat *at, struct svc_req *rqstp)
{
  abaddrproc_stat stat;

  switch ((stat=abaddrproc(rqstp))) {
  case ABADDRPROC_OK: break;
  default: return stat; break;
  }
  switch (at->af) {
  case AF_INET:
    at->ads[0].in.s_addr = htonl(at->ads[0].in.s_addr);
    at->ads[1].in.s_addr = htonl(at->ads[1].in.s_addr);
	/* host byte order at xdr of rpc, and convert to network byte order */
    break;
  }
  return stat;
}

abaddrproc_stat abaddrproc_addr_set(struct abaddrat *at, struct svc_req *rqstp)
{
  abaddrproc_stat stat;

  switch ((stat=abaddrproc_addr(at, rqstp))) {
  case ABADDRPROC_OK: break;
  default: return stat; break;
  }
  if (!abaddrat_db_del(&me, at))
    stat = ABADDRPROC_EIO;
  if (!(at->lt<0))
    if (!abaddrat_db_add(&me, at))
      stat = ABADDRPROC_EIO;
  abaddr_db_sync(&me);
  return stat;
}

abaddrproc_stat *
abaddrproc_req_serv_1_svc(abaddr_req *argp, struct svc_req *rqstp)
{
  static abaddrproc_stat stat;
  int check = 0;

#ifdef DEBUG
  check = !0;
#endif
  switch ((stat=abaddrproc(rqstp))) {
  case ABADDRPROC_OK: break;
  default: return &stat; break;
  }
  switch (*argp) {
  case ABADDR_REQ_DB_LIFE:
    abaddr_db_cycle_life(&me);
    break;
  case ABADDR_REQ_DB_TRNC:
    abaddr_db_terminate(&me);
    abaddr_db_prepare(&me);
    break;
  case ABADDR_REQ_DB_LOAD:
    abaddr_db_amset_read(&me,
			 &amset,
#if !defined(DISABLE_IPv6)
			 &am6set,
#endif
			 !0, check);
    break;
  case ABADDR_REQ_DB_SAVE:
    abaddr_db_amset_write(&me);
    break;
  }
  return &stat;
}

abaddrproc_stat *
abaddrproc_addr_set_1_svc(abaddrattr *argp, struct svc_req *rqstp)
{
  static abaddrproc_stat stat;
  struct abaddrat at = { 0, };

  switch (argp->addr.af) {
  case ABADDRF_INET:
    at.af = AF_INET;
    at.ads[0].in = *(struct in_addr *)&argp->addr.abaddrs_u.s[0];
    at.ads[1].in = *(struct in_addr *)&argp->addr.abaddrs_u.s[1];
    break;
#if !defined(DISABLE_IPv6)
  case ABADDRF_INET6:
    at.af = AF_INET6;
    at.ads[0].in6 = *(struct in6_addr *)&argp->addr.abaddrs_u.s6[0];
    at.ads[1].in6 = *(struct in6_addr *)&argp->addr.abaddrs_u.s6[16];
    break;
#endif
  default:
    return NULL;
    break;
  }
  at.rc = argp->rc;
  at.lt = argp->lt;
  strlcpy(at.c, argp->c, sizeof(at.c));
  stat = abaddrproc_addr_set(&at, rqstp);
  return &stat;
}

abaddrproc_stat abaddrproc_addr_get(struct abaddrat *at, struct svc_req *rqstp)
{
  abaddrproc_stat stat;

  switch ((stat=abaddrproc_addr(at, rqstp))) {
  case ABADDRPROC_OK: break;
  default: return stat; break;
  }
  if (!abaddrat_db_get(&me, at))
    at->af = AF_UNSPEC;
  return stat;
}

abaddr_get_stat *
abaddrproc_addr_get_1_svc(abaddrattr *argp, struct svc_req *rqstp)
{
  static abaddr_get_stat stat = { 0, };
  static struct abaddrat at = { 0, };

  switch (argp->addr.af) {
  case ABADDRF_INET:
    at.af = AF_INET;
    at.ads[0].in = *(struct in_addr *)&argp->addr.abaddrs_u.s[0];
    at.ads[1].in = *(struct in_addr *)&argp->addr.abaddrs_u.s[1];
    break;
#if !defined(DISABLE_IPv6)
  case ABADDRF_INET6:
    at.af = AF_INET6;
    at.ads[0].in6 = *(struct in6_addr *)&argp->addr.abaddrs_u.s6[0];
    at.ads[1].in6 = *(struct in6_addr *)&argp->addr.abaddrs_u.s6[16];
    break;
#endif
  default:
    return NULL;
    break;
  }
  at.rc = argp->rc;
  at.lt = argp->lt;
  strlcpy(at.c, argp->c, sizeof(at.c));
  switch ((stat.stat=abaddrproc_addr_get(&at, rqstp))) {
  case ABADDRPROC_OK: break;
  default: return &stat; break;
  }
  switch (at.af) {
  case AF_INET:
    stat.at.addr.af = ABADDRF_INET;
    at.ads[0].in.s_addr = ntohl(at.ads[0].in.s_addr);
    at.ads[1].in.s_addr = ntohl(at.ads[1].in.s_addr);
	/* host byte order at xdr of rpc, and convert from network byte order */
    *(struct in_addr *)&stat.at.addr.abaddrs_u.s[0] = at.ads[0].in;
    *(struct in_addr *)&stat.at.addr.abaddrs_u.s[1] = at.ads[1].in;
    break;
#if !defined(DISABLE_IPv6)
  case AF_INET6:
    stat.at.addr.af = ABADDRF_INET6;
    *(struct in6_addr *)&stat.at.addr.abaddrs_u.s6[0] = at.ads[0].in6;
    *(struct in6_addr *)&stat.at.addr.abaddrs_u.s6[16] = at.ads[1].in6;
    break;
#endif
  default:
    stat.at.addr.af = 0;
    break;
  }
  stat.at.rc = at.rc;
  stat.at.lt = at.lt;
  stat.at.c = at.c;
  return &stat;
}

#ifndef no_abaddrd_main
int main(int argc, char *argv[])
{
  int abaddr_main();
  int a, check = 0;

#ifdef DEBUG
  check = !0;
#endif
  me.argc = argc;
  me.argv = argv;
  for (a=1; a<argc; a++)
    if (strcmp(argv[a], "--") == 0)
      break;
    else if (strcmp(argv[a], "-C") == 0)
      me.clear = !0;
    else if (strcmp(argv[a], "-R") == 0)
      me.read_lst = !0;
    else if (strcmp(argv[a], "-W") == 0)
      me.write_lst = !0;
    else if (strcmp(argv[a], "-w") == 0 && a+1 < argc)
      me.wt = strtoul(argv[++a], NULL, 10);
    else if (strcmp(argv[a], "-f") == 0 && a+1 < argc)
      me.pathname = argv[++a];
  am_init_masks();
#if !defined(DISABLE_IPv6)
  am6_init_masks();
#endif
  amset_clients_read(&amset_clients,
#if !defined(DISABLE_IPv6)
		     &am6set_clients,
#endif
		     0, check, argc, argv);
  abaddr_db_prepare(&me);
  if (me.read_lst)
    abaddr_db_amset_read(&me,
			 &amset,
#if !defined(DISABLE_IPv6)
			 &am6set,
#endif
			 !0, check);
  if (me.write_lst) abaddr_db_amset_write(&me);
  abaddr_main(argc, argv);
  return 1;
}
#endif
