/*
 * abcidr.c - utilities for CIDR in abmail
 *
 * Copyright (C) 2007-2008 Taiji Yamada <taiji@aihara.co.jp>
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Affero General Public License for more details.
 */
/*
 * usage:
 *	$ ./abcidr [-c-] 192.168.0.0-192.168.2.255 192.168.3.0/23 ...
 */
#include <stdio.h>	/* stderr, etc */
#include <string.h>	/* strcmp, etc */
#include <stdlib.h>	/* qsort, bsearch, etc */
#include <sys/types.h>	/* netinet/in.h, arpa/inet.h, sys/socket.h, etc */
#include <sys/socket.h>	/* AF_INET6, etc */
#include <netinet/in.h>	/* in_addr, etc */
#include <arpa/inet.h>	/* inet_ntoa, etc */

#define min(a, b) (((a)<(b))?(a):(b))
#define max(a, b) (((a)>(b))?(a):(b))
#define lbu(l, b, u) min(max((l), (b)), (u))

#if !defined(HAVE_STRLCPY)
#define strlcpy(d, s, z) do { strncpy(d, s, z); if ((z) > 0) (d)[(z)-1] = '\0'; } while (0)
#endif

#if !defined(HAVE_STRLCAT)
#define strlcat(d, s, z) do { if ((z) > strlen(d)) strncat(d, s, (z)-strlen(d)-1); } while (0)
#endif

#if !defined(HAVE_INET_ATON)
#define inet_aton(s, p) ((((p)->s_addr=inet_addr(s))!=-1||strcmp(s, "255.255.255.255")==0)?1:0)
#endif

struct am {
  char a[16], m[1+16];
  int rc;
  char *c;
  struct in_addr ads[4];	/* 0: address, 1: mask, 2: network, 3: broadcast */
};

int am_compare(const void *a, const void *b)
{
  struct am *am_a = (struct am *)a, *am_b = (struct am *)b;

  if (ntohl(am_b->ads[2].s_addr) <= ntohl(am_a->ads[3].s_addr) &&
      ntohl(am_a->ads[2].s_addr) <= ntohl(am_b->ads[3].s_addr))
    return 0;
  else if (ntohl(am_a->ads[3].s_addr) < ntohl(am_b->ads[2].s_addr))
    return -1;
  else
    return 1;
}

struct in_addr am_masks[32+1];

void am_init_masks(void)
{
  int l;

  for (l=0; l<sizeof(am_masks)/sizeof(struct in_addr); l++) {
    am_masks[l].s_addr = 0;
    if (l > 0)
      am_masks[l].s_addr = am_masks[l-1].s_addr | htonl(0x80000000>>(l-1));
  }
}

int am_set_space(struct am *am)
{
  int l, lm = -1;

  if (am->m[0] != '-') {
    if (am->m[0] == '\0')
      am->ads[1] = am_masks[(lm=32)];
    else if (!strchr(am->m, '.'))
      am->ads[1] = am_masks[(lm=lbu(0, atoi(am->m), 32))];
    else if (inet_aton(am->m, &am->ads[1]) == 0)
      fprintf(stderr, "inet_aton failed for %s\n", am->m);
    else
      for (l=0; l<sizeof(am_masks)/sizeof(struct in_addr); l++)
	if (am->ads[1].s_addr == am_masks[l].s_addr) {
	  lm = l;
	  break;
	}
    am->ads[2].s_addr = am->ads[0].s_addr&am->ads[1].s_addr;
    am->ads[3].s_addr = am->ads[0].s_addr|~am->ads[1].s_addr;
  }
  else {
    am->ads[1] = am_masks[0];
    am->ads[2] = am->ads[0];
    if (inet_aton(&am->m[1], &am->ads[3]) == 0)
      fprintf(stderr, "inet_aton failed for %s\n", &am->m[1]);
  }
  return lm;
}

int am_put_cidr(struct am *am, int lm)
{
  strlcpy(am->a, inet_ntoa(am->ads[0]), sizeof(am->a));
  sprintf(am->m, "%d", (lm=lbu(0, lm, 32)));	/* not insecure */
  return lm;
}

int am_put_mask(struct am *am)
{
  int l, lm = -1;

  for (l=0; l<sizeof(am_masks)/sizeof(struct in_addr); l++)
    if (am->ads[1].s_addr == am_masks[l].s_addr) {
      lm = l;
      break;
    }
  if (lm == -1) {
    strlcpy(am->a, inet_ntoa(am->ads[0]), sizeof(am->a));
    strlcpy(am->m, inet_ntoa(am->ads[1]), sizeof(am->m));
  }
  else
    am_put_cidr(am, lm);
  return lm;
}

int am_put_range(struct am *am)
{
  strlcpy(am->a, inet_ntoa(am->ads[2]), sizeof(am->a));
  strlcpy(am->m, "-", sizeof(am->m));
  strlcat(am->m, inet_ntoa(am->ads[3]), sizeof(am->m));
  return 0;
}

int am_compress(int n, struct am ams[])
{
  int i;

  for (i=1; i<n;)
    if (ams[i-1].rc == ams[i].rc &&
	htonl(ntohl(ams[i-1].ads[3].s_addr) + 1) == ams[i].ads[2].s_addr) {
      struct am a = ams[i];

      ams[i-1].ads[3].s_addr = ams[i].ads[3].s_addr;
      ams[i-1].ads[1] = am_masks[0];
      am_put_range(&ams[i-1]);
      if (i+1 < n) {
	memmove(&ams[i], &ams[i+1], sizeof(struct am)*(n-2));
	ams[n-1] = a;
      }
      n--;
    }
    else
      i++;
  return n;
}

int am_fprint_space(struct am *am, FILE *fp)
{
  if (fp) fprintf(fp, "%s%s%s", am->a, (am->m[0] != '-')?"/":"", am->m);
  return 0;
}

int am_fprint_range(struct am *am, FILE *fp)
{
  if (fp) fprintf(fp, "%s-", inet_ntoa(am->ads[2]));
  if (fp) fprintf(fp, "%s", inet_ntoa(am->ads[3]));
  return 0;
}

int am_fprint_cidrs(struct am *am, FILE *fp)
{
  struct am a = *am;
  int l, lm, m = 0, n = 0;

  do {
    struct in_addr lux;
    int ln;

    if ((lux.s_addr = a.ads[2].s_addr^a.ads[3].s_addr) == 0)
      m = sizeof(am_masks)/sizeof(struct in_addr)-1;
    else
      for (; m<sizeof(am_masks)/sizeof(struct in_addr)-1; m++)
	if (htonl(ntohl(lux.s_addr)<<m)&am_masks[1].s_addr)
	  break;
    for (lm=ln=-1, l=m; l<sizeof(am_masks)/sizeof(struct in_addr); l++) {
      struct in_addr lbm, lbb, ubm;

      lbm.s_addr = a.ads[2].s_addr&am_masks[l].s_addr;
      lbb.s_addr = a.ads[2].s_addr|~am_masks[l].s_addr;
      ubm.s_addr = a.ads[3].s_addr&am_masks[l].s_addr;
      if (lbm.s_addr == a.ads[2].s_addr) {
	if (lbb.s_addr == a.ads[3].s_addr) {
	  lm = l;
	  break;
	}
	if (ntohl(lbb.s_addr) <= ntohl(a.ads[3].s_addr) &&
	    ntohl(ubm.s_addr) <= ntohl(a.ads[3].s_addr)) {
	  ln = l;
	  break;
	}
      }
    }
    if (!(lm < 0)) {
      a.ads[1] = am_masks[lm];
      am_put_cidr(&a, lm);
      am_fprint_space(&a, fp);
    }
    else {
      struct am b = a;

      b.ads[0] = a.ads[2];
      b.ads[1] = am_masks[ln];
      b.ads[2].s_addr = b.ads[0].s_addr&b.ads[1].s_addr;
      b.ads[3].s_addr = b.ads[0].s_addr|~b.ads[1].s_addr;
      am_put_cidr(&b, ln);
      am_fprint_space(&b, fp);
      if (fp) fprintf(fp, ",");
      a.ads[2].s_addr = htonl(ntohl(b.ads[3].s_addr) + 1);
      a.ads[0] = a.ads[2];
    }
    n++;
  } while (lm < 0);
  return n;
}

int am_fprint_ranges(struct am *am, FILE *fp)
{
  struct am a = *am;
  struct in_addr rm, m, v, w, x;
  int i, l, u, n = 0, indice[sizeof(am_masks)/sizeof(struct in_addr)-1];

  rm.s_addr = ~a.ads[1].s_addr;
  for (l=0; l<sizeof(am_masks)/sizeof(struct in_addr)-1; l++)
    if (htonl(ntohl(rm.s_addr)<<l)&am_masks[1].s_addr)
      break;
  for (u=l; u<sizeof(am_masks)/sizeof(struct in_addr)-1; u++)
    if ((rm.s_addr&(m.s_addr=~am_masks[u].s_addr)) == m.s_addr)
      break;
  x.s_addr = ~(rm.s_addr&am_masks[u].s_addr);
  for (i=l; i<u; i++)
    if (htonl(ntohl(rm.s_addr)<<i)&am_masks[1].s_addr)
      indice[n++] = i;
  for (v.s_addr=0; v.s_addr<(1<<n); v.s_addr++) {
    struct am b = a;

    w.s_addr = 0;
    for (i=0; i<n; i++)
      w.s_addr |= (v.s_addr<<(32-1-i)&ntohl(am_masks[1].s_addr))>>indice[n-1-i];
    b.ads[0].s_addr = (a.ads[0].s_addr&x.s_addr)|htonl(w.s_addr);
    b.ads[1] = am_masks[u];
    am_put_cidr(&b, u);
    am_set_space(&b);
    am_fprint_range(&b, fp);
    if (fp) fprintf(fp, (v.s_addr!=(1<<n)-1)?",":"");
  }
  return 1<<n;
}

#if !defined(DISABLE_IPv6)
#include "arr_ar.h"
struct in6_addr in6_addr_zero(void)
{
  struct in6_addr a;

  arr_init(16, a.s6_addr);
  return a;
}
struct in6_addr in6_addr_one(void)
{
  struct in6_addr a;

  arr_init(16, a.s6_addr);
  a.s6_addr[16-1] = 1;
  return a;
}
#if 0
struct in6_addr in6_addr_not(struct in6_addr a)
{
  struct in6_addr b;

  arr_not(16, a.s6_addr, b.s6_addr);
  return b;
}
struct in6_addr in6_addr_and(struct in6_addr a, struct in6_addr b)
{
  struct in6_addr c;

  arr_and(16, a.s6_addr, b.s6_addr, c.s6_addr);
  return c;
}
struct in6_addr in6_addr_or(struct in6_addr a, struct in6_addr b)
{
  struct in6_addr c;

  arr_or(16, a.s6_addr, b.s6_addr, c.s6_addr);
  return c;
}
struct in6_addr in6_addr_xor(struct in6_addr a, struct in6_addr b)
{
  struct in6_addr c;

  arr_xor(16, a.s6_addr, b.s6_addr, c.s6_addr);
  return c;
}
struct in6_addr in6_addr_lsr(struct in6_addr a, unsigned int s)
{
  struct in6_addr b;

  arr_lsr(16, a.s6_addr, s, b.s6_addr);
  return b;
}
struct in6_addr in6_addr_lsl(struct in6_addr a, unsigned int s)
{
  struct in6_addr b;

  arr_lsl(16, a.s6_addr, s, b.s6_addr);
  return b;
}
int in6_addr_cmp(struct in6_addr a, struct in6_addr b)
{
  return memcmp(&a.s6_addr[0], &b.s6_addr[0], sizeof(struct in6_addr));
}
struct in6_addr in6_addr_add(struct in6_addr a, struct in6_addr b)
{
  struct in6_addr c;

  arr_add(16, a.s6_addr, b.s6_addr, c.s6_addr);
  return c;
}
struct in6_addr in6_addr_sub(struct in6_addr a, struct in6_addr b)
{
  struct in6_addr c;

  arr_sub(16, a.s6_addr, b.s6_addr, c.s6_addr);
  return c;
}
int in6_addr_istrue(struct in6_addr a)
{
  int i;

  arr_is(16, a.s6_addr[i]);
  return i<16;
}
#else
struct in6_addr in6_addr_not(struct in6_addr a)
{
  struct in6_addr b;

  arr_not(4, (uint32_t *)a.s6_addr, (uint32_t *)b.s6_addr);
  return b;
}
struct in6_addr in6_addr_and(struct in6_addr a, struct in6_addr b)
{
  struct in6_addr c;

  arr_and(4, (uint32_t *)a.s6_addr, (uint32_t *)b.s6_addr, (uint32_t *)c.s6_addr);
  return c;
}
struct in6_addr in6_addr_or(struct in6_addr a, struct in6_addr b)
{
  struct in6_addr c;

  arr_or(4, (uint32_t *)a.s6_addr, (uint32_t *)b.s6_addr, (uint32_t *)c.s6_addr);
  return c;
}
struct in6_addr in6_addr_xor(struct in6_addr a, struct in6_addr b)
{
  struct in6_addr c;

  arr_xor(4, (uint32_t *)a.s6_addr, (uint32_t *)b.s6_addr, (uint32_t *)c.s6_addr);
  return c;
}
struct in6_addr in6_addr_lsr(struct in6_addr a, unsigned int s)
{
  struct in6_addr b;

#if 1
  arr_lsr(16, a.s6_addr, s, b.s6_addr);
#else
  arr_tr(4, (uint32_t *)a.s6_addr, ntohl, (uint32_t *)a.s6_addr);
  arr_lsr(4, (uint32_t *)a.s6_addr, s, (uint32_t *)b.s6_addr);
  arr_tr(4, (uint32_t *)b.s6_addr, htonl, (uint32_t *)b.s6_addr);
#endif
  return b;
}
struct in6_addr in6_addr_lsl(struct in6_addr a, unsigned int s)
{
  struct in6_addr b;

  arr_tr(4, (uint32_t *)a.s6_addr, ntohl, (uint32_t *)a.s6_addr);
  arr_lsl(4, (uint32_t *)a.s6_addr, s, (uint32_t *)b.s6_addr);
  arr_tr(4, (uint32_t *)b.s6_addr, htonl, (uint32_t *)b.s6_addr);
  return b;
}
int in6_addr_cmp(struct in6_addr a, struct in6_addr b)
{
#if 1
  return memcmp(&a.s6_addr[0], &b.s6_addr[0], sizeof(struct in6_addr));
#else
  int i, cmp;

  arr_tr(4, (uint32_t *)a.s6_addr, ntohl, (uint32_t *)a.s6_addr);
  arr_tr(4, (uint32_t *)b.s6_addr, ntohl, (uint32_t *)b.s6_addr);
  arr_is(4, (cmp=((uint32_t *)a.s6_addr)[i]>((uint32_t *)b.s6_addr)[i]?1:
	     ((uint32_t *)a.s6_addr)[i]<((uint32_t *)b.s6_addr)[i]?-1:0) != 0);
  return cmp;
#endif
}
struct in6_addr in6_addr_add(struct in6_addr a, struct in6_addr b)
{
  struct in6_addr c;

  arr_tr(4, (uint32_t *)a.s6_addr, ntohl, (uint32_t *)a.s6_addr);
  arr_tr(4, (uint32_t *)b.s6_addr, ntohl, (uint32_t *)b.s6_addr);
  arr_add(4, (uint32_t *)a.s6_addr, (uint32_t *)b.s6_addr, (uint32_t *)c.s6_addr);
  arr_tr(4, (uint32_t *)c.s6_addr, htonl, (uint32_t *)c.s6_addr);
  return c;
}
struct in6_addr in6_addr_sub(struct in6_addr a, struct in6_addr b)
{
  struct in6_addr c;

  arr_tr(4, (uint32_t *)a.s6_addr, ntohl, (uint32_t *)a.s6_addr);
  arr_tr(4, (uint32_t *)b.s6_addr, ntohl, (uint32_t *)b.s6_addr);
  arr_sub(4, (uint32_t *)a.s6_addr, (uint32_t *)b.s6_addr, (uint32_t *)c.s6_addr);
  arr_tr(4, (uint32_t *)c.s6_addr, htonl, (uint32_t *)c.s6_addr);
  return c;
}
int in6_addr_istrue(struct in6_addr a)
{
  int i;

  arr_is(4, ((uint32_t *)a.s6_addr)[i]);
  return i<4;
}
#endif
#define in6_addr_eq(a, b) (in6_addr_cmp(a, b) == 0)
#define in6_addr_ne(a, b) (in6_addr_cmp(a, b) != 0)
#define in6_addr_lt(a, b) (in6_addr_cmp(a, b) < 0)
#define in6_addr_gt(a, b) (in6_addr_cmp(a, b) > 0)
#define in6_addr_le(a, b) (in6_addr_cmp(a, b) <= 0)
#define in6_addr_ge(a, b) (in6_addr_cmp(a, b) >= 0)

struct am6 {
  char a[64], m[1+64];
  int rc;
  char *c;
  struct in6_addr ads[4];	/* 0: address, 1: mask, 2: network, 3: netend */
};

int am6_compare(const void *a, const void *b)
{
  struct am6 *am_a = (struct am6 *)a, *am_b = (struct am6 *)b;

  if (in6_addr_le(am_b->ads[2], am_a->ads[3]) &&
      in6_addr_le(am_a->ads[2], am_b->ads[3]))
    return 0;
  else if (in6_addr_lt(am_a->ads[3], am_b->ads[2]))
    return -1;
  else
    return 1;
}

struct in6_addr am6_masks[128+1];

void am6_init_masks(void)
{
  int l;
  struct in6_addr msb = in6_addr_zero();

  msb.s6_addr[0] = 0x80;
  for (l=0; l<sizeof(am6_masks)/sizeof(struct in6_addr); l++) {
    am6_masks[l] = in6_addr_zero();
    if (l > 0)
      am6_masks[l] = in6_addr_or(am6_masks[l-1], in6_addr_lsr(msb, l-1));
  }
}

int am6_set_space(struct am6 *am)
{
  int l, lm = -1;

  if (am->m[0] != '-') {
    if (am->m[0] == '\0')
      am->ads[1] = am6_masks[(lm=128)];
    else if (!strchr(am->m, ':') && !strchr(am->m, '.'))
      am->ads[1] = am6_masks[(lm=lbu(0, atoi(am->m), 128))];
    else if (inet_pton(AF_INET6, am->m, &am->ads[1]) == 0)
      fprintf(stderr, "inet_pton failed for %s\n", am->m);
    else
      for (l=0; l<sizeof(am6_masks)/sizeof(struct in6_addr); l++)
	if (in6_addr_eq(am->ads[1], am6_masks[l])) {
	  lm = l;
	  break;
	}
    am->ads[2] = in6_addr_and(am->ads[0], am->ads[1]);
    am->ads[3] = in6_addr_or(am->ads[0], in6_addr_not(am->ads[1]));
  }
  else {
    am->ads[1] = am6_masks[0];
    am->ads[2] = am->ads[0];
    if (inet_pton(AF_INET6, &am->m[1], &am->ads[3]) == 0)
      fprintf(stderr, "inet_pton failed for %s\n", &am->m[1]);
  }
  return lm;
}

int am6_put_cidr(struct am6 *am, int lm)
{
  inet_ntop(AF_INET6, &am->ads[0], am->a, sizeof(am->a));
  sprintf(am->m, "%d", (lm=lbu(0, lm, 128)));	/* not insecure */
  return lm;
}

int am6_put_mask(struct am6 *am)
{
  int l, lm = -1;

  for (l=0; l<sizeof(am6_masks)/sizeof(struct in6_addr); l++)
    if (in6_addr_eq(am->ads[1], am6_masks[l])) {
      lm = l;
      break;
    }
  if (lm == -1) {
    inet_ntop(AF_INET6, &am->ads[0], am->a, sizeof(am->a));
    inet_ntop(AF_INET6, &am->ads[1], am->m, sizeof(am->m));
  }
  else
    am6_put_cidr(am, lm);
  return lm;
}

int am6_put_range(struct am6 *am)
{
  inet_ntop(AF_INET6, &am->ads[2], am->a, sizeof(am->a));
  strlcpy(am->m, "-", sizeof(am->m));
  inet_ntop(AF_INET6, &am->ads[3], &am->m[1], sizeof(am->m)-1);
  return 0;
}

int am6_compress(int n, struct am6 ams[])
{
  int i;
  struct in6_addr one = in6_addr_one();

  for (i=1; i<n;)
    if (ams[i-1].rc == ams[i].rc &&
	in6_addr_eq(in6_addr_add(ams[i-1].ads[3], one), ams[i].ads[2])) {
      struct am6 a = ams[i];

      ams[i-1].ads[3] = ams[i].ads[3];
      ams[i-1].ads[1] = am6_masks[0];
      am6_put_range(&ams[i-1]);
      if (i+1 < n) {
	memmove(&ams[i], &ams[i+1], sizeof(struct am6)*(n-2));
	ams[n-1] = a;
      }
      n--;
    }
    else
      i++;
  return n;
}

int am6_fprint_space(struct am6 *am, FILE *fp)
{
  if (fp) fprintf(fp, "%s%s%s", am->a, (am->m[0] != '-')?"/":"", am->m);
  return 0;
}

int am6_fprint_range(struct am6 *am, FILE *fp)
{
  char s[INET6_ADDRSTRLEN];

  if (fp) fprintf(fp, "%s-", inet_ntop(AF_INET6, &am->ads[2], s, sizeof(s)));
  if (fp) fprintf(fp, "%s", inet_ntop(AF_INET6, &am->ads[3], s, sizeof(s)));
  return 0;
}

int am6_fprint_cidrs(struct am6 *am, FILE *fp)
{
  struct am6 a = *am;
  int l, lm, m = 0, n = 0;
  struct in6_addr one = in6_addr_one();

  do {
    struct in6_addr lux;
    int ln;

    if (!in6_addr_istrue(lux = in6_addr_xor(a.ads[2], a.ads[3])))
      m = sizeof(am6_masks)/sizeof(struct in6_addr)-1;
    else
      for (; m<sizeof(am6_masks)/sizeof(struct in6_addr)-1; m++)
	if (in6_addr_istrue(in6_addr_and(in6_addr_lsl(lux, m), am6_masks[1])))
	  break;
    for (lm=ln=-1, l=m; l<sizeof(am6_masks)/sizeof(struct in6_addr); l++) {
      struct in6_addr lbm, lbb, ubm;

      lbm = in6_addr_and(a.ads[2], am6_masks[l]);
      lbb = in6_addr_or(a.ads[2], in6_addr_not(am6_masks[l]));
      ubm = in6_addr_and(a.ads[3], am6_masks[l]);
      if (in6_addr_eq(lbm, a.ads[2])) {
	if (in6_addr_eq(lbb, a.ads[3])) {
	  lm = l;
	  break;
	}
	if (in6_addr_le(lbb, a.ads[3]) &&
	    in6_addr_le(ubm, a.ads[3])) {
	  ln = l;
	  break;
	}
      }
    }
    if (!(lm < 0)) {
      a.ads[1] = am6_masks[lm];
      am6_put_cidr(&a, lm);
      am6_fprint_space(&a, fp);
    }
    else {
      struct am6 b = a;

      b.ads[0] = a.ads[2];
      b.ads[1] = am6_masks[ln];
      b.ads[2] = in6_addr_and(b.ads[0], b.ads[1]);
      b.ads[3] = in6_addr_or(b.ads[0], in6_addr_not(b.ads[1]));
      am6_put_cidr(&b, ln);
      am6_fprint_space(&b, fp);
      if (fp) fprintf(fp, ",");
      a.ads[2] = in6_addr_add(b.ads[3], one);
      a.ads[0] = a.ads[2];
    }
    n++;
  } while (lm < 0);
  return n;
}

int am6_fprint_ranges(struct am6 *am, FILE *fp)
{
  struct am6 a = *am;
  struct in6_addr rm, m, v, w, x, one = in6_addr_one(), nv;
  int i, l, u, n = 0, indice[sizeof(am6_masks)/sizeof(struct in6_addr)-1];

  rm = in6_addr_not(a.ads[1]);
  for (l=0; l<sizeof(am6_masks)/sizeof(struct in6_addr)-1; l++)
    if (in6_addr_istrue(in6_addr_and(in6_addr_lsl(rm, l), am6_masks[1])))
      break;
  for (u=l; u<sizeof(am6_masks)/sizeof(struct in6_addr)-1; u++)
    if (in6_addr_eq(in6_addr_and(rm, (m=in6_addr_not(am6_masks[u]))), m))
      break;
  x = in6_addr_not(in6_addr_and(rm, am6_masks[u]));
  for (i=l; i<u; i++)
    if (in6_addr_istrue(in6_addr_and(in6_addr_lsl(rm, i), am6_masks[1])))
      indice[n++] = i;
  for (v=in6_addr_zero(), nv=in6_addr_lsl(one, n); in6_addr_lt(v, nv); v=in6_addr_add(v, one)) {
    struct am6 b = a;

    w = in6_addr_zero();
    for (i=0; i<n; i++)
      w = in6_addr_or(w, in6_addr_lsr(in6_addr_and(in6_addr_lsl(v, 128-1-i), am6_masks[1]), indice[n-1-i]));
    b.ads[0] = in6_addr_or(in6_addr_and(a.ads[0], x), w);
    b.ads[1] = am6_masks[u];
    am6_put_cidr(&b, u);
    am6_set_space(&b);
    am6_fprint_range(&b, fp);
    if (fp) fprintf(fp, in6_addr_ne(v, in6_addr_sub(in6_addr_lsl(one, n), one))?",":"");
  }
  return 1<<n;
}
#endif

int main(int argc, char *argv[])
{
  int a, i, n;
  int compress = !0;
  char *p, *q, *r;
  struct am am = { "", "", 0, NULL, }, *ams = NULL;
#if !defined(DISABLE_IPv6)
  int n6;
  struct am6 am6 = { "", "", 0, NULL, }, *am6s = NULL;
#endif

  for (a=1; a<argc; a++)
    if (strcmp(argv[a], "--") == 0)
      break;
    else if (strcmp(argv[a], "-c-") == 0)
      compress = 0;
    else
      break;
  am_init_masks();
#if !defined(DISABLE_IPv6)
  am6_init_masks();
#endif
  for (; a<argc; a++) {
    int a_compress = 0, a_ot = -1;
#if !defined(DISABLE_IPv6)
#define am_set_space(a)		(af==AF_INET6)? am6_set_space(a##6):		(am_set_space)(a)
#define am_fprint_range(a,fp)	(af==AF_INET6)? am6_fprint_range(a##6,fp):	(am_fprint_range)(a,fp)
#define am_fprint_ranges(a,fp)	(af==AF_INET6)? am6_fprint_ranges(a##6,fp):	(am_fprint_ranges)(a,fp)
#define am_fprint_cidrs(a,fp)	(af==AF_INET6)? am6_fprint_cidrs(a##6,fp):	(am_fprint_cidrs)(a,fp)
    int af = AF_INET;

    n6 = 0;
#endif
    n = 0;
    p = argv[a];
    while (p) {
      int l;

      if ((r = strchr(p, ',')))
	*r = '\0';
      if (r && !a_compress)
	a_compress = compress;
      if ((q = strchr(p, '/'))) {
	*q = '\0';
	q++;
	strlcpy(am.a, p, sizeof(am.a));
	strlcpy(am.m, q, sizeof(am.m));
	if (inet_aton(p, &am.ads[0]) == 0) {
#if !defined(DISABLE_IPv6)
	  if (inet_pton(AF_INET6, p, &am6.ads[0]) == 0)
	    fprintf(stderr, "inet_pton failed for %s\n", p);
	  else {
	    strlcpy(am6.a, p, sizeof(am6.a));
	    strlcpy(am6.m, q, sizeof(am6.m));
	    af = AF_INET6;
	  }
#else
	  fprintf(stderr, "inet_aton failed for %s\n", p);
#endif
	}
	if ((l = am_set_space(&am)) != -1) {
	  if (!a_compress) am_fprint_range(&am, stdout);
	  else if (a_ot == -1) a_ot = 0;
	}
	else {
	  if (!a_compress) am_fprint_ranges(&am, stdout);
	  else if (a_ot == -1) a_ot = 2;
	}
      }
      else if ((q = strchr(p, '-'))) {
	strlcpy(am.m, q, sizeof(am.m));
	*q = '\0';
	strlcpy(am.a, p, sizeof(am.a));
	if (inet_aton(p, &am.ads[0]) == 0) {
#if !defined(DISABLE_IPv6)
	  if (inet_pton(AF_INET6, p, &am6.ads[0]) == 0)
	    fprintf(stderr, "inet_pton failed for %s\n", p);
	  else {
	    strlcpy(am6.a, p, sizeof(am6.a));
	    strlcpy(am6.m, "-", sizeof(am6.m));
	    strlcat(am6.m, q+1, sizeof(am6.m));
	    af = AF_INET6;
	  }
#else
	  fprintf(stderr, "inet_aton failed for %s\n", p);
#endif
	}
	l = am_set_space(&am);
	if (!a_compress) am_fprint_cidrs(&am, stdout);
	else if (a_ot == -1) a_ot = 1;
      }
      if (!a_compress) fprintf(stdout, r?",":"\n");
#if !defined(DISABLE_IPv6)
      else if (af == AF_INET6) {
	am6s = am6s?realloc(am6s, sizeof(struct am6)*(n6+1)):
	  malloc(sizeof(struct am6)*(n6+1));
	am6s[n6++] = am6;
      }
#endif
      else {
	ams = ams?realloc(ams, sizeof(struct am)*(n+1)):
	  malloc(sizeof(struct am)*(n+1));
	ams[n++] = am;
      }
      p = r?(r + 1):NULL;
    }
#if !defined(DISABLE_IPv6)
    if (am6s) {
      qsort(am6s, n6, sizeof(struct am6), am6_compare);
      n6 = am6_compress(n6, am6s);
      for (i=0; i<n6; i++) {
	am6 = am6s[i];
	switch (a_ot) {
	case 0:		am_fprint_range(&am, stdout);	break;
	case 1:		am_fprint_cidrs(&am, stdout);	break;
	default:	am_fprint_ranges(&am, stdout);	break;
	}
	fprintf(stdout, i!=n6-1?",":"\n");
      }
      free(am6s);
      am6s = NULL;
    }
#endif
    if (ams) {
      qsort(ams, n, sizeof(struct am), am_compare);
      n = am_compress(n, ams);
      for (i=0; i<n; i++) {
	am = ams[i];
	switch (a_ot) {
	case 0:		am_fprint_range(&am, stdout);	break;
	case 1:		am_fprint_cidrs(&am, stdout);	break;
	default:	am_fprint_ranges(&am, stdout);	break;
	}
	fprintf(stdout, i!=n-1?",":"\n");
      }
      free(ams);
      ams = NULL;
    }
  }
  return 0;
}
